import re

from typing import List, Optional


SPACE = " "
COMMENT_CHAR = "-"
MULTI_LINE_COMMENT_START = ["/", "*"]
MULTI_LINE_COMMENT_END = ["*", "/"]
NEWLINE = "\n"
WHITESPACE_REGEX = re.compile(r"\s+")
NEED_SPACE_AROUND_CHARS = [";", ",", ".", "(", ")", "+", "-", "*", "/", "="]
QUOTES = {"\"", "'"}


def get_tokens(source: str) -> List[str]:
    """
    Converts a raw source string to a list of tokens.
    - Strips whitespace from either side.
    - Adds spaces around characters like commas so they become tokens.
    - Ignores extra whitespace or line breaks in the query.
    """
    normalized = source.strip()
    cleaned = WHITESPACE_REGEX.sub(SPACE, normalized)
    for char in NEED_SPACE_AROUND_CHARS:
        cleaned = cleaned.replace(char, SPACE + char + SPACE)
    separated = cleaned.split(SPACE)
    tokens = [t for t in separated if t]
    return tokens


def is_single_line_comment_at(i: int, chars: List[str]) -> bool:
    """Checks if a single line comment starts at position `i` of `chars`."""
    if len(chars) <= i + 1:
        return False
    return chars[i] == COMMENT_CHAR and chars[i + 1] == COMMENT_CHAR


def is_multi_line_comment_at(i: int, chars: List[str]) -> bool:
    """Checks if a multi line comment starts at position `i` of `chars`."""
    if len(chars) <= i + 1:
        return False
    return (
        chars[i] == MULTI_LINE_COMMENT_START[0]
        and chars[i + 1] == MULTI_LINE_COMMENT_START[1]
    )


def is_end_of_multi_line_comment_at(i: int, chars: List[str]) -> bool:
    """Checks if a multi line comment ends at position `i` of `chars`."""
    if len(chars) <= i + 1:
        return False
    return (
        chars[i] == MULTI_LINE_COMMENT_END[0]
        and chars[i + 1] == MULTI_LINE_COMMENT_END[1]
    )


def get_tokens_string_safe(source: str) -> List[str]:
    """
    Converts a raw source string to a list of tokens.
    - Strips whitespace from either side.
    - Ensures that non-whitespace separating characters become single tokens.
    - Creates a single token for a quoted string, even if contains whitespace or
      separating characters.
    - Ignores comments.
    - Ignores other extra whitespace or line breaks in the query.
    """
    chars: str = source.strip()
    # Scan characters to accumulate tokens
    tokens: List[str] = []
    token_chars: List[str] = []
    open_quote: Optional[str] = None
    ignore_next: bool = False
    is_single_line_comment: bool = False
    is_multi_line_comment: bool = False
    for i, c in enumerate(chars):
        if ignore_next:
            ignore_next = False
            pass
        elif is_single_line_comment:
            if c == NEWLINE:
                # Found end of single line comment
                is_single_line_comment = False
            else:
                # Found char in single line comment
                pass
        elif is_multi_line_comment:
            if is_end_of_multi_line_comment_at(i, chars):
                # Found end of multi line comment
                is_multi_line_comment = False
                ignore_next = True
            else:
                # Found char in multi line comment
                pass
        elif open_quote is not None:
            if c == open_quote:
                # Found close quote
                token_chars.append(c)
                token = "".join(token_chars)
                tokens.append(token)
                token_chars = []
                open_quote = None
            else:
                # Found char in quoted string
                token_chars.append(c)
        elif WHITESPACE_REGEX.match(c):
            # Found non-quoted whitespace
            if token_chars:
                token = "".join(token_chars)
                tokens.append(token)
                token_chars = []
        elif is_single_line_comment_at(i, chars):
            # Found start of single line comment
            is_single_line_comment = True
        elif is_multi_line_comment_at(i, chars):
            # Found start of multi line comment
            is_multi_line_comment = True
        elif c in NEED_SPACE_AROUND_CHARS:
            # Found character token
            if token_chars:
                token = "".join(token_chars)
                tokens.append(token)
                token_chars = []
            tokens.append(c)
        elif c in QUOTES:
            # Found open quote
            if token_chars:
                token = "".join(token_chars)
                tokens.append(token)
                token_chars = []
            open_quote = c
            token_chars.append(c)
        else:
            # Found other char
            token_chars.append(c)
    # Capture all remaining accumulated token chars into a token
    if token_chars:
        token = "".join(token_chars)
        tokens.append(token)
        token_chars = []
    return tokens



from .types import ZqlQuery, SqlQuery
from .parser import AstParseError
from .loader import get_zql_grammar
from .renderer import QueryRenderError
from .translator import translate
from .types import MaybeDialect


ZQL_DIALECT = "zql"
TARGET_DIALECT = "sqlite"
ZQL_GRAMMAR = get_zql_grammar()


class ZqlParserError(Exception):
    pass


def get_dialect_from_name(name: str) -> MaybeDialect:
    if name == ZQL_DIALECT:
        return None
    return name


class Zql:
    """Converts ZQL queries to SQL."""

    def __init__(self):
        pass

    def parse(self, raw: ZqlQuery) -> SqlQuery:
        try:
            sql = translate(ZQL_GRAMMAR, raw, target_dialect=TARGET_DIALECT)
            return sql
        except AstParseError as ape:
            raise ZqlParserError(ape)
        except QueryRenderError as qre:
            raise ZqlParserError(ape)

    def translate(self, raw: str, source: str, target: str) -> str:
        source_dialect = get_dialect_from_name(source)
        target_dialect = get_dialect_from_name(target)
        try:
            sql = translate(ZQL_GRAMMAR, raw, source_dialect, target_dialect)
            return sql
        except AstParseError as ape:
            raise ZqlParserError(ape)
        except QueryRenderError as qre:
            raise ZqlParserError(ape)


import re

from typing import Dict, List

from .types import MaybeDialect


DEF = ":"
CASE = "|"
TEMPLATE = ">"
DIALECT = "<"
END = ";"
REGEX_START = "@r"
QUOTE = "\""
SPACE = " "
COMMA = ","
NEWLINE = "\n"
ESCAPED_NEWLINE = "\\n"
TEMPLATES_KEY = "templates"
ROOT = "root"


Grammar = Dict[str, List[dict]]


class GrammarParseError(Exception):
    pass


def parse_rule(rule: str) -> dict:
    is_template = rule.startswith(TEMPLATE)
    is_regex_rule = rule.startswith(REGEX_START)
    is_literal_rule = rule.startswith(QUOTE)

    dialect_index = rule.rfind(DIALECT)
    close_quote_index = rule.rfind(QUOTE)

    has_no_dialect_symbol = dialect_index < 0
    has_dialect_symbol_in_quotes = (
        (is_literal_rule or is_template)
        and close_quote_index > dialect_index
    )
    has_dialects = (
        not has_no_dialect_symbol
        and not has_dialect_symbol_in_quotes
    )

    dialects: List[str] = []
    rule_content = rule
    if has_dialects:
        raw_dialects = rule[dialect_index + 2:]
        dialects = [d.strip() for d in raw_dialects.split(COMMA)]
        rule_content = rule[:dialect_index - 1]


    parsed_rule: dict = {}
    if is_regex_rule:
        regex = rule_content[2:]
        parsed_rule["regex"] = regex
    elif is_literal_rule:
        literal = rule_content[1:-1]
        parsed_rule["literal"] = literal
    elif is_template:
        template_no_quotes = rule_content[3:-1]
        template = template_no_quotes.replace(ESCAPED_NEWLINE, NEWLINE)
        parsed_rule["template"] = template
    else:
        nodes = rule_content.split(SPACE)
        parsed_rule["sequence"] = nodes

    if dialects:
        parsed_rule["dialects"] = dialects

    return parsed_rule


def parse_grammar(content: str) -> Grammar:
    raw_lines = [l.strip() for l in content.split("\n")]

    current_node = None
    grammar = {}
    for i, raw_line in enumerate(raw_lines):
        n = i + 1
        line = raw_line.strip()
        if not line:
            continue

        if line == END:
            current_node = None
            continue

        if not current_node:
            def_index = line.find(DEF)
            if def_index < 0:
                raise GrammarParseError(f"L{n}: Missing definition `:`.")

            node = line[:def_index].strip()
            if node in grammar:
                raise GrammarParseError(f"L{n}: Repeat definition of `{node}`.")

            grammar[node] = []
            current_node = node

            raw_rule = line[def_index+1:].strip()
            if not raw_rule:
                raise GrammarParseError(f"L{n}: Missing rule after `:`.")

            rule = parse_rule(raw_rule)
            grammar[current_node].append(rule)
            continue

        template_index = line.find(TEMPLATE)
        case_index = line.find(CASE)

        has_template = template_index > -1
        has_case = case_index > -1
        has_template_first = template_index < case_index
        is_template = has_template and (not has_case or has_template_first)
        if is_template:
            raw_template = line[template_index:].strip()
            if not raw_template:
                raise GrammarParseError(f"L{n}: Missing template after `>`.")

            template = parse_rule(raw_template)
            current_rules = grammar[current_node]
            if not current_rules:
                raise GrammarParseError(
                    f"L{n}: Missing rule for template on lines above `>`."
                )

            last_rule = current_rules[-1]
            if TEMPLATES_KEY not in last_rule:
                last_rule[TEMPLATES_KEY] = []

            last_rule[TEMPLATES_KEY].append(template)
            continue

        if case_index < 0:
            raise GrammarParseError(f"L{n}: Missing case `|`.")

        raw_rule = line[case_index+1:].strip()
        if not raw_rule:
            raise GrammarParseError(f"L{n}: Missing rule after `|`.")

        rule = parse_rule(raw_rule)
        grammar[current_node].append(rule)

    if current_node:
        raise GrammarParseError("Missing `;` to end grammar.")

    if ROOT not in grammar:
        raise GrammarParseError("Missing `root` definition.")

    return grammar


def is_relevant_to_dialect(obj: dict, dialect: MaybeDialect) -> bool:
    obj_dialects: List[str] = obj.get("dialects", [])
    is_default_dialect = dialect is None and not obj_dialects
    has_dialect = dialect is not None and dialect in obj_dialects
    supports_any_dialect = not obj_dialects
    is_relevant = (
        is_default_dialect
        or has_dialect
        or supports_any_dialect
    )
    return is_relevant


from . import cleaner
from . import grammar
from . import loader
from . import main
from . import parser
from . import renderer
from . import translator
from . import types

from .types import ZqlQuery
from .main import Zql, ZqlParserError


from typing import Dict, List, Optional, Tuple

from .grammar import Grammar, is_relevant_to_dialect
from .parser import AstNode
from .types import SqlQuery, MaybeDialect


RuleKey = Tuple[str, List[str]]
Template = str
TemplateLookup = Dict[RuleKey, Template]


SPACE = " "
SEQUENCE_RULE_TYPE = "sequence"
NON_CHILDREN_RULE_TYPES = {"literal", "regex"}


class QueryRenderError(Exception):
    pass


def render_query(
    grammar: Grammar,
    ast: AstNode,
    target_dialect: MaybeDialect = None
) -> SqlQuery:
    template_lookup = get_template_lookup(grammar, target_dialect)
    return render_with_grammar(grammar, template_lookup, ast)


def render_with_grammar(
    grammar: Grammar,
    lookup: TemplateLookup,
    ast: AstNode
) -> SqlQuery:
    node_type = ast.get("type")
    children = ast.get("children", [])
    if not node_type:
        raise QueryRenderError(f"Node should have a `type`: {ast}")

    template = maybe_get_template(lookup, ast)
    if template:
        kwargs = {
            c.get("type"): render_with_grammar(grammar, lookup, c)
            for c in children
        }
        rendered = template.format(**kwargs)
        return rendered

    if children:
        args = [render_with_grammar(grammar, lookup, c) for c in children]
        rendered = SPACE.join(args)
        return rendered

    value = ast.get("value")
    if value is not None:
        return value

    raise QueryRenderError(f"Unable to render node: `{node_type}`.")


def maybe_get_template_for_dialect(
    rule: dict,
    target_dialect: MaybeDialect
) -> Optional[str]:
    templates: List[dict] = rule.get("templates")
    if templates is None:
        return None
    
    for template in templates:
        if is_relevant_to_dialect(template, target_dialect):
            return template.get("template")

    return None
    

def get_template_lookup(
    grammar: Grammar,
    target_dialect: MaybeDialect
) -> TemplateLookup:
    template_lookup: TemplateLookup = {}
    for node, rules in grammar.items():
        for rule in rules:
            template = maybe_get_template_for_dialect(rule, target_dialect)
            rule_dialects: List[str] = rule.get("dialects", [])
            literal = rule.get("literal")
            sequence = rule.get("sequence")

            implicit_template = None
            if literal:
                implicit_template = literal
            if sequence:
                nodes = ["{" + name + "}" for name in sequence]
                implicit_template = SPACE.join(nodes)

            no_template = template is None and implicit_template is not None
            if no_template and is_relevant_to_dialect(rule, target_dialect):
                template = implicit_template

            if template is None:
                continue

            key = get_rule_key(node, rule)
            if key in template_lookup:
                break

            template_lookup[key] = template

    return template_lookup


def get_rule_key(node: str, rule: dict) -> RuleKey:
    if SEQUENCE_RULE_TYPE in rule:
        sequence = rule.get(SEQUENCE_RULE_TYPE)
        key = SPACE.join([node, *sequence])
        return key

    for rule_type in NON_CHILDREN_RULE_TYPES:
        if rule_type in rule:
            key = SPACE.join([node, rule_type])
            return key

    raise QueryRenderError(f"Unable to determine pattern of node: `{node}`.")


def maybe_get_template(
    lookup: TemplateLookup,
    ast: AstNode
) -> Optional[Template]:
    node = ast.get("type")
    children = ast.get("children", [])

    if children:
        rule_pattern = [child.get("type") for child in children]
        key = SPACE.join([node, *rule_pattern])
        template = lookup.get(key)
        return template

    for rule_type in NON_CHILDREN_RULE_TYPES:
        key = SPACE.join([node, rule_type])
        template = lookup.get(key)
        if template is not None:
            return template

    return None



from pathlib import Path
from .grammar import Grammar, parse_grammar


ZQL_GRAMMAR_PATH = Path(__file__).parent / "zql_grammar.tmjd"


def get_zql_grammar() -> Grammar:
    with ZQL_GRAMMAR_PATH.open("r") as file:
        content = file.read()
        return parse_grammar(content)


from .grammar import Grammar
from .parser import parse_ast
from .renderer import render_query
from .types import MaybeDialect


def translate(
    grammar: Grammar,
    source: str,
    source_dialect: MaybeDialect = None,
    target_dialect: MaybeDialect = None
) -> str:
    ast = parse_ast(grammar, source, source_dialect)
    target = render_query(grammar, ast, target_dialect)
    return target


import re

from typing import List

from .grammar import ROOT, Grammar, is_relevant_to_dialect
from .cleaner import get_tokens_string_safe
from .types import MaybeDialect


SPACE = " "

AstNode = dict


class AstParseError(Exception):
    pass


class TokensManager:

    def __init__(self, tokens: List[str]):
        self.tokens = tokens

    def set_tokens(self, tokens: List[str]):
        self.tokens = tokens

    def copy(self):
        return TokensManager(list(self.tokens))


def evaluate_literal(tokens: List[str], literal: str) -> AstNode:
    tokens_in_literal = len(literal.split(SPACE))
    peeked_tokens = SPACE.join(tokens[:tokens_in_literal]).casefold()
    if peeked_tokens != literal.casefold():
        raise AstParseError(f"Expected `{literal}`. Got `{peeked_tokens}`.")
    
    for _ in range(tokens_in_literal):
        tokens.pop(0)
    return {"value": literal}


def evaluate_regex(tokens: List[str], regex: str) -> AstNode:
    if not tokens:
        raise AstParseError(f"Expected match for `{regex}`, not end of input.")

    next_token = tokens[0]
    if not re.compile(regex).match(next_token):
        raise AstParseError(f"Expected `{next_token}` to match `{regex}`.")

    tokens.pop(0)
    return {"value": next_token}


def evaluate_sequence(
    grammar: Grammar,
    tokens_manager: TokensManager,
    sequence: List[str],
    source_dialect: MaybeDialect
) -> AstNode:
    mutable_tokens_manager = tokens_manager.copy()
    children: List[AstNode] = []
    for node in sequence:
        ast_node = evaluate_node(
            grammar,
            mutable_tokens_manager,
            node,
            source_dialect
        )
        children.append(ast_node)

    tokens_manager.set_tokens(mutable_tokens_manager.tokens)
    return {"children": children}


def evaluate_rule(
    grammar: Grammar,
    tokens_manager: TokensManager,
    rule: dict,
    source_dialect: MaybeDialect
) -> AstNode:
    tokens = tokens_manager.tokens
    literal = rule.get("literal")
    if literal is not None:
        ast_node = evaluate_literal(tokens, literal)
        return ast_node
    
    regex = rule.get("regex")
    if regex is not None:
        ast_node = evaluate_regex(tokens, regex)
        return ast_node
    
    sequence = rule.get("sequence")
    if sequence is not None:
        mutable_tokens_manager = tokens_manager.copy()
        ast_node = evaluate_sequence(
            grammar,
            mutable_tokens_manager,
            sequence,
            source_dialect
        )
        mutated_tokens = mutable_tokens_manager.tokens
        tokens_manager.set_tokens(mutated_tokens)
        return ast_node

    raise AstParseError(f"Invalid rule: {rule}")


def evaluate_node(
    grammar: Grammar,
    tokens_manager: TokensManager,
    node: str,
    source_dialect: MaybeDialect
) -> AstNode:
    rules = grammar.get(node, [])
    if not rules:
        raise AstParseError(
            f"Reached node `{node}`, which has no defined rules."
        )

    error = None
    ast_node = None
    for rule in rules:
        if not is_relevant_to_dialect(rule, source_dialect):
            continue

        try:
            mutable_tokens_manager = tokens_manager.copy()
            rule_node = evaluate_rule(
                grammar,
                mutable_tokens_manager,
                rule,
                source_dialect
            )

            remaining_tokens = mutable_tokens_manager.tokens
            if node == ROOT and remaining_tokens:
                raise AstParseError(
                    "Could not apply `root` rule to remaining tokens: "
                    f"{remaining_tokens}"
                )

            tokens_manager.set_tokens(remaining_tokens)
            ast_node = rule_node
            break
        except Exception as e:
            error = e
            continue

    if not ast_node:
        if error:
            raise error

        tokens = tokens_manager.tokens
        remaining_source_sample = SPACE.join(tokens[:3])[:20]
        raise AstParseError(
            f"Failed to parse `{node}` at: `{remaining_source_sample}`."
        )

    ast_node = {"type": node, **ast_node}
    return ast_node


def parse_ast(
    grammar: Grammar,
    source: str,
    source_dialect: MaybeDialect = None
) -> AstNode:
    tokens = get_tokens_string_safe(source)
    tokens_manager = TokensManager(tokens)
    root = evaluate_node(grammar, tokens_manager, ROOT, source_dialect)

    remaining_tokens = tokens_manager.tokens
    if remaining_tokens:
        raise AstParseError(
            "Satisfied `root` rule, but unparsed tokens remain: "
            f"{remaining_tokens}"
        )

    children = root.get("children")
    if not children:
        raise AstParseError("Did not parse anything for `root`.")

    n = len(children)
    if n > 1:
        raise AstParseError(f"Parsed {n} nodes for `root`. Expected only one.")

    return children[0]



from enum import Enum
from typing import Dict, Optional


ZqlQuery = str
SqlQuery = str
AstNode = Dict
Token = str
MaybeDialect = Optional[str]


class NodeType(Enum):
    QUERY = "query"
    SELECT = "select"
    FROM = "from"
    WHERE = "where"
    LIMIT = "limit"
    FILTER = "filter"
    FILTER_BRANCH = "filter_branch"
    EXPRESSION = "expression"
    INTEGER = "integer"
    TERMINAL = "terminal"
